from typing import Literal

from peropq.pauli import PauliString


def get_commutator_pauli_tensors(
    left_pauli_string: PauliString,
    right_pauli_string: PauliString,
) -> PauliString | Literal[0]:
    """Calculate the commutator of any two pauli tensors.

    :param left_pauli_string left side of commutator
    :param right_pauli_string right side of commutator.

    The general formula for A = c(A_1 x ... A_n), B = d(B_1 x ... x B_N)
    is [A, B] = 1 - (-1)^k cd(A_1 B_1 x ... x A_N B_N)
    where x is the tensor product, and k is the number of
    anti-commuting Pauli pairs.

    :returns: None if the strings commute, otherwise the commutator as a PauliString
    """
    if left_pauli_string.commutes_with(right_pauli_string):
        return 0
    return 2 * left_pauli_string * right_pauli_string



from collections.abc import Sequence

from peropq import pauli


class Hamiltonian:
    """Class representing the Hamilonian."""

    def __init__(self, pauli_string_list: Sequence[pauli.PauliString]) -> None:
        """
        Initialization.

        param: pauli_string_list contains all the terms of the hamiltonian.
        """
        self.pauli_string_list = pauli_string_list
        # Initialize the Pauli strings with coeff 1.0 and store the coefficients in self.cjs.
        cjs: list[complex] = []
        for pauli_string in self.pauli_string_list:
            cjs.append(pauli_string.coefficient)
            pauli_string.coefficient = 1.0
        self.cjs = cjs

    def get_n_terms(self) -> int:
        """Method returning the number of Pauli Strings."""
        return len(self.pauli_string_list)

    def get_cjs(self) -> list[complex]:
        """Method returning the coefficients."""
        return self.cjs



from collections.abc import Sequence

import numpy as np
import numpy.typing as npt
import scipy  # type: ignore[import-untyped]

from peropq.variational_unitary import VariationalUnitary


class Optimizer:
    """Class performing the optimizer."""

    def optimize(
        self,
        variational_unitary: VariationalUnitary,
        initial_guess: Sequence[float] = [],
    ) -> tuple[scipy.optimize.OptimizeResult, float]:
        """
        Perform the minimization.

        param: variational_unitary ansatz used for optimization
        param: initial_guess initial guess for the optimization. If not provided, use the parameters of the Trotterization instead
        returns: the result of the optimization
        returns: the perturbative 2-norm
        """
        if len(initial_guess) != 0:
            x0: npt.NDArray = np.array(initial_guess)
        else:
            x0 = variational_unitary.get_initial_trotter_vector()
            x0 = variational_unitary.flatten_theta(x0)
        if not variational_unitary.trace_calculated:
            variational_unitary.calculate_traces()
        optimized_results = scipy.optimize.minimize(variational_unitary.c2_squared, x0)
        return optimized_results, variational_unitary.c2_squared(
            theta=optimized_results.x,
        )



import scipy
import scipy.sparse as sp  # type: ignore[import-untyped]
from numpy import typing as npt

from peropq.hamiltonian import Hamiltonian
from peropq.pauli import Pauli, PauliString
from peropq.variational_unitary import VariationalUnitary


class ExactDiagonalization:
    """Class implementing exact diagonalization for benchmarks on small system sizes."""

    def __init__(self, number_of_qubits: int) -> None:
        """
        Initializing the ed module.

        param: qubit_number number of qubits.
        """
        self.number_of_qubits = number_of_qubits

    def pauli_to_sparse(self, pauli: Pauli) -> sp.spmatrix:
        """
        Converts a Pauli to a 2x2 sparse matrix.

        param: pauli to be converted.
        """
        match pauli:
            case Pauli.I:
                return sp.csc_matrix([[1.0, 0], [0, 1.0]], dtype=complex)
            case Pauli.X:
                return sp.csc_matrix([[0, 1.0], [1.0, 0]], dtype=complex)
            case Pauli.Y:
                return sp.csc_matrix([[0, -1j], [1j, 0]], dtype=complex)
            case Pauli.Z:
                return sp.csc_matrix([[1.0, 0.0], [0.0, -1.0]], dtype=complex)

    def get_sparse(self, pauli_string: PauliString) -> sp.spmatrix:
        """
        Transforms PauliString into sparse matrix.

        param: pauli_string to be transformed into a sparse matrix.
        """
        sparse_string: sp.spmatrix = pauli_string.coefficient * self.pauli_to_sparse(
            pauli_string.get_pauli(0),
        )
        for qubit in range(1, self.number_of_qubits):
            sparse_string = sp.kron(
                sparse_string,
                self.pauli_to_sparse(pauli_string.get_pauli(qubit)),
            )
        return sp.csc_matrix(sparse_string)

    def get_hamiltonian_matrix(self, hamiltonian: Hamiltonian) -> sp.spmatrix:
        """param: hamiltonian to be converted to sparse."""
        hamiltonian_matrix = self.get_sparse(hamiltonian.pauli_string_list[0])
        for i_string in range(1, len(hamiltonian.pauli_string_list)):
            hamiltonian_matrix += self.get_sparse(
                hamiltonian.pauli_string_list[i_string],
            )
        return hamiltonian_matrix

    def get_continuous_time_evolution(
        self,
        hamiltonian: Hamiltonian,
        time: float,
    ) -> sp.spmatrix:
        """
        Get the continuous time evolution of an Hamiltonian.

        param: hamiltonian governing the dynamics
        param: time at which we want to time evolve.
        """
        hamiltonian_matrix = self.get_hamiltonian_matrix(hamiltonian)
        return sp.linalg.expm(-1j * time * hamiltonian_matrix)

    def get_variational_evolution(
        self,
        variational_unitary: VariationalUnitary,
    ) -> sp.spmatrix:
        """
        Get the time evolution unitary from a variational unitary.

        param: variational_unitary to be evolved.
        """
        u_sparse = sp.eye(2**self.number_of_qubits)
        for layer in range(variational_unitary.depth):
            for i_term, term in enumerate(variational_unitary.pauli_string_list):
                u_sparse = u_sparse @ sp.linalg.expm(
                    -1j
                    * variational_unitary.theta[layer, i_term]
                    * self.get_sparse(term),
                )
        return u_sparse

    def apply_continuous_to_state(
        self,
        hamiltonian: Hamiltonian,
        time: float,
        state: npt.NDArray,
    ) -> npt.NDArray:
        """
        Apply the continuous time evolution.

        param: hamiltonian governing the dynamics
        param: time at which we want to time evolve.
        param: state to be evolved.
        """
        hamiltonian_matrix = self.get_hamiltonian_matrix(hamiltonian)
        return scipy.sparse.linalg.expm_multiply(
            -1j * time * hamiltonian_matrix,
            state,
        )

    def apply_variational_to_state(
        self,
        variational_unitary: VariationalUnitary,
        state: npt.NDArray,
    ) -> npt.NDArray:
        """
        Apply the variational unitary to a state.

        param: variational_unitary to be applied
        param: state on which the unitary is applied
        """
        unitary = self.get_variational_evolution(variational_unitary)
        return unitary @ state

    def get_error(
        self,
        variational_unitary: VariationalUnitary,
        hamiltonian: Hamiltonian,
    ) -> float:
        """
        Returns the error made for by the variational unitary (compared to the continuous time evolution).

        param: variational_unitary to be compared with the continuous time evolution.
        param: hamiltonian used to caclulated the continuous time evolution.
        """
        sparse_variational_unitary = self.get_variational_evolution(variational_unitary)
        sparse_continuous_unitary = self.get_continuous_time_evolution(
            hamiltonian,
            time=variational_unitary.time,
        )
        return sp.linalg.norm(
            sparse_continuous_unitary - sparse_variational_unitary,
            ord="fro",
        )



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Sequence

import numpy as np
import numpy.typing as npt
from scipy.sparse import csr_array  # type: ignore[import-untyped]

from peropq.commutators import get_commutator_pauli_tensors
from peropq.hamiltonian import Hamiltonian
from peropq.pauli import PauliString


class VariationalUnitary:
    """class representing the variational unitary ansataz."""

    def __init__(
        self,
        hamiltonian: Hamiltonian,
        number_of_layer: int,
        time: float,
    ) -> None:
        """
        Init function.

        :param hamiltonian Hamiltonian of which one is interested in the dynamics.
        :param R number of steps for the optimization.
        :param t final time to up to which one wants to perform the time evolution.
        """
        self.n_terms: int = hamiltonian.get_n_terms()
        self.pauli_string_list: Sequence[PauliString] = hamiltonian.pauli_string_list
        self.depth: int = number_of_layer
        self.theta: npt.NDArray = np.zeros((number_of_layer, self.n_terms))
        self.cjs: Sequence[complex] = hamiltonian.get_cjs()
        self.time: float = time
        self.test: npt.NDArray = np.zeros((number_of_layer, number_of_layer))
        self.trace_calculated = False
        for r in range(number_of_layer):
            for s in range(number_of_layer):
                self.test[r, s] = -1 if s > r else 1

    def update_theta(self, new_array: npt.NDArray) -> None:
        """
         Update theta ensuring that the condition Sum_i theta_i dt_i= is ensured.

        :param new_array the new array containing the variational parameters. It's shape must be (R - 1, n_terms).
        """
        if new_array.shape != (self.depth - 1, self.n_terms):
            error_message = "Wrong length provided."
            raise ValueError(error_message)
        for j in range(self.n_terms):
            for r in range(self.depth - 1):
                self.theta[r, j] = new_array[r, j]
            self.theta[self.depth - 1, j] = self.time * self.cjs[j]
            for r in range(self.depth - 1):
                self.theta[self.depth - 1, j] -= new_array[r, j]

    def get_initial_trotter_vector(self) -> npt.NDArray:
        """Get the variational parameters corresponding to the Trotterization. Useful to initialize the optimization."""
        theta_trotter: npt.NDArray = np.zeros((self.depth - 1, self.n_terms))
        for j in range(self.n_terms):
            for r in range(self.depth - 1):
                theta_trotter[r, j] = self.cjs[j] * self.time / self.depth
        return theta_trotter

    def flatten_theta(self, theta: npt.NDArray) -> npt.NDArray:
        """Returns the variational parameters as flatten (R-1)*n_terms array. Useful to pass to a minimization function."""
        return np.array(theta).reshape((self.depth - 1) * self.n_terms)

    def set_theta_to_trotter(self) -> None:
        """Sets the variational parameters to the Trotter parameters."""
        theta_trotter: npt.NDArray = self.get_initial_trotter_vector()
        self.update_theta(theta_trotter)

    def chi_tensor(
        self,
        left_indices: npt.NDArray,
        right_indices: npt.NDArray,
    ) -> npt.NDArray:
        """
        Vectorized function to calculate all the chi coefficient at once.

        param: left_indices indices of the left tensor which give non-zero contributions in the calculation of chi.
        param: right_indices indices of the right tensor which give non-zero contributions in the calculation of chi.
        """
        theta_left: npt.NDArray = self.theta[:, left_indices]
        theta_right: npt.NDArray = self.theta[:, right_indices]
        res: npt.NDArray = np.tensordot(theta_left, self.test, ([0], [0]))
        res = np.tensordot(res, theta_right, ([1], [0]))
        return 0.5 * res

    def calculate_traces(self) -> None:
        """Calculate the trace tensor."""
        commutators: list[tuple[int, PauliString]] = []
        index_pairs: list[tuple[int, int]] = []
        i = 0
        for j_prime, h_j_prime in enumerate(
            self.pauli_string_list,
        ):
            for j in range(j_prime + 1, self.n_terms):
                h_j = self.pauli_string_list[j]
                commutator = get_commutator_pauli_tensors(h_j, h_j_prime)
                if commutator:
                    index_pairs.append((j, j_prime))
                    commutators.append((i, commutator))
                    i += 1
        self.index_pairs: npt.NDArray = np.array(index_pairs)
        self.left_indices: npt.NDArray = np.unique(self.index_pairs[:, 0])
        self.right_indices: npt.NDArray = np.unique(self.index_pairs[:, 1])
        self.trace_tensor: npt.NDArray = 1j * np.zeros(
            (
                len(self.left_indices),
                len(self.right_indices),
                len(self.left_indices),
                len(self.right_indices),
            ),
        )
        for j_, j_commutator in commutators:
            for k_, k_commutator in commutators:
                if j_ < k_:
                    continue
                product_commutators: PauliString = j_commutator * k_commutator
                if product_commutators != 0:
                    trace: complex = product_commutators.normalized_trace()
                    if trace:
                        fac = 1 if j_ == k_ else 2.0
                        # Get the new indices
                        new_j: int = np.where(
                            self.left_indices == self.index_pairs[j_, 0],
                        )[0].item()
                        new_j_prime: int = np.where(
                            self.right_indices == self.index_pairs[j_, 1],
                        )[0].item()
                        new_k: int = np.where(
                            self.left_indices == self.index_pairs[k_, 0],
                        )[0].item()
                        new_k_prime: int = np.where(
                            self.right_indices == self.index_pairs[k_, 1],
                        )[0].item()
                        self.trace_tensor[new_j, new_j_prime, new_k, new_k_prime] = (
                            fac * product_commutators.normalized_trace()
                        )
        self.trace_calculated = True

    def c2_squared(self, theta: npt.ArrayLike = ()) -> float:
        """
        Perturbative 2-norm.

        param: theta parameters of the variational unitary.
        returns: the perturbative approximation of the 2-norm difference between the exact and the variational representation.
        """
        if not self.trace_calculated:
            self.calculate_traces()
        if np.array(theta).shape[0] != 0:
            theta_new = np.array(theta).reshape(
                (
                    self.depth - 1,
                    self.n_terms,
                ),
            )
            self.update_theta(theta_new)
        chi_tensor = self.chi_tensor(
            self.left_indices,
            self.right_indices,
        )
        s1, s2, s3, s4 = self.trace_tensor.shape
        chi_tensor = chi_tensor.reshape((s1 * s2,))
        trace_tensor: csr_array = csr_array(
            self.trace_tensor.reshape((s1 * s2, s3 * s4)),
        )
        return np.real(-chi_tensor.T @ trace_tensor @ chi_tensor)



from __future__ import annotations

from cmath import isclose
from dataclasses import dataclass, field
from enum import Enum
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Sequence


class Pauli(Enum):
    """Enum class for Pauli gates."""

    I = "I"
    X = "X"
    Y = "Y"
    Z = "Z"

    def __mul__(self, other: Pauli) -> tuple[complex, Pauli]:
        """Multiply two Paulis."""
        match (self, other):
            case (Pauli.X, Pauli.Y) | (Pauli.Y, Pauli.X):
                coeff = 1j if self == Pauli.X else -1j
                return coeff, Pauli.Z
            case (Pauli.Y, Pauli.Z) | (Pauli.Z, Pauli.Y):
                coeff = 1j if self == Pauli.Y else -1j
                return coeff, Pauli.X
            case (Pauli.Z, Pauli.X) | (Pauli.X, Pauli.Z):
                coeff = 1j if self == Pauli.Z else -1j
                return coeff, Pauli.Y
            case (Pauli.I, _):
                return 1, other
            case (_, Pauli.I):
                return 1, self
        return 1, Pauli.I

    def commutes_with(self, other: Pauli) -> bool:
        """Return whether this Pauli commutes with given Pauli.

        :param other The right side of the Commutator

        :return: True if PauliStrings commute, False otherwise
        """
        if self == other:
            return True
        if self == Pauli.I:
            return True
        if other == Pauli.I:
            return True
        return False


@dataclass()
class PauliString:
    """Class representing a Pauli string multiplied by a complex coefficient."""

    qubit_pauli_map: dict[int, Pauli] = field(default_factory=dict)
    coefficient: complex = 1

    @classmethod
    def from_pauli_sequence(
        cls,
        paulis: Sequence[Pauli],
        coeff: complex = 1,
        start_qubit: int = 0,
    ) -> PauliString:
        """Create from a sequence of Paulis.

        The Paulis will be placed sequentially on qubits starting from qubit index <start_qubit>

        :param paulis a list/tuple of Paulis
        :param coeff coefficient of the tensor
        :param start_qubit which qubit to place first Pauli on
        """
        return cls({(i + start_qubit): pauli for i, pauli in enumerate(paulis)}, coeff)

    def __post_init__(self) -> None:
        """Special dataclass method that runs directly after initialization."""
        self.prune()

    def prune(self) -> PauliString:
        """Remove any Identities from the string."""
        qubits = list(self.qubit_pauli_map.keys())
        for qubit in qubits:
            if self.get_pauli(qubit) == Pauli.I:
                self.qubit_pauli_map.pop(qubit)
        return self

    def __eq__(self, other: object) -> bool:
        """Equality for PauliStrings.

        Two PauliStrings are equal if their coefficients are equal and
        if their Paulis are equal on all qubits. The underlying qubit_pauli_maps
        may differ (in the number of Identities).
        """
        if not isinstance(other, PauliString):
            return False
        if not isclose(self.coefficient, other.coefficient):
            return False
        leftover_qubits = set(other.qubit_pauli_map.keys())
        for qubit, self_pauli in self.qubit_pauli_map.items():
            if self_pauli != other.get_pauli(qubit):
                return False
            leftover_qubits.discard(qubit)
        return all(other.get_pauli(qubit) == Pauli.I for qubit in leftover_qubits)

    def get_pauli(self, qubit: int) -> Pauli:
        """Return the pauli at a given qubit.

        :param qubit the qubit index to retrieve the Pauli from

        :returns: the pauli at qubit index <qubit> otherwise, the identity Pauli.I
        """
        if qubit < 0:
            message = "qubit index cannot be negative"
            raise ValueError(message)
        return self.qubit_pauli_map.get(qubit, Pauli.I)

    def update(
        self,
        update_qubit_pauli_map: dict[int, Pauli],
        coeff: complex | None = None,
    ) -> PauliString:
        """Update the PauliString."""
        for qubit, pauli in update_qubit_pauli_map.items():
            if pauli == Pauli.I:
                self.qubit_pauli_map.pop(qubit, Pauli.I)
            else:
                self.qubit_pauli_map[qubit] = pauli
        if coeff:
            self.coefficient = self.coefficient * coeff
        return self

    def __mul__(self, other: PauliString | complex) -> PauliString:
        """Multiply PauliString with a complex number or PauliString."""
        if isinstance(other, PauliString):
            return _pauli_string_mult(self, other)
        return PauliString(self.qubit_pauli_map, self.coefficient * other)

    def __rmul__(self, other: complex) -> PauliString:
        """Multiply PauliString with a complex number or PauliString."""
        return PauliString(self.qubit_pauli_map, self.coefficient * other)

    def commutes_with(self, other: PauliString) -> bool:
        """Return whether this PauliString commutes with given PauliString.

        :param other The right side of the Commutator

        :return: True if PauliStrings commute, False otherwise
        """
        number_anti_commute = 0
        for qubit, self_pauli in self.qubit_pauli_map.items():
            other_pauli = other.get_pauli(qubit)
            if not self_pauli.commutes_with(other_pauli):
                number_anti_commute += 1
        if number_anti_commute % 2 == 0:
            return True
        return False

    def normalized_trace(self) -> complex:
        """Returns `2^(-n)*Tr(self)`.

        The trace is just the coefficient times the product of the traces of the Paulis
        over all n qubits (including any identities on qubits not explicitly contained in the qubit_pauli_map). Since
        `Tr(Pauli.I) == 2`, and `Tr(Pauli.X) == Tr(Pauli.Y) == Tr(Pauli.Z) == 0`, the result will either be 0 (if any
        of the Paulis is not the Identity) or `self.coefficient` (if all paulis are the identity).
        """
        self.prune()
        if len(self.qubit_pauli_map) == 0:
            return self.coefficient
        return 0


def _pauli_string_mult(
    pauli_string1: PauliString,
    pauli_string2: PauliString,
) -> PauliString:
    new_pauli_string = PauliString(
        {},
        pauli_string1.coefficient * pauli_string2.coefficient,
    )
    leftover_qubits = set(pauli_string2.qubit_pauli_map.keys())
    for qubit, left_pauli in pauli_string1.qubit_pauli_map.items():
        leftover_qubits.discard(qubit)
        right_pauli = pauli_string2.get_pauli(qubit)
        coeff, new_pauli = left_pauli * right_pauli
        new_pauli_string.update({qubit: new_pauli}, coeff)
    for qubit in leftover_qubits:
        right_pauli = pauli_string2.get_pauli(qubit)
        if right_pauli != Pauli.I:
            # left pauli must be I
            new_pauli_string.update({qubit: right_pauli})
    return new_pauli_string
